from django.core.exceptions import ObjectDoesNotExist
from django_readers.utils import map_or_apply
from operator import attrgetter, methodcaller


def wrap(key, value_getter):
    def projector(instance):
        return {key: value_getter(instance)}

    return projector


def attr(name, *, transform_value=None, transform_value_if_none=False):
    def value_getter(instance):
        value = attrgetter(name)(instance)
        if transform_value and (value is not None or transform_value_if_none):
            value = transform_value(value)
        return value

    return wrap(name, value_getter)


def method(name, *args, **kwargs):
    return wrap(name, methodcaller(name, *args, **kwargs))


def relationship(name, related_projector):
    """
    Given an attribute name and a projector, return a projector which plucks
    the attribute off the instance, figures out whether it represents a single
    object or an iterable/queryset of objects, and applies the given projector
    to the related object or objects.
    """

    def value_getter(instance):
        try:
            related = attrgetter(name)(instance)
        except ObjectDoesNotExist:
            return None
        return map_or_apply(related, related_projector)

    return wrap(name, value_getter)


def pk_list(name):
    """
    Given an attribute name (which should be a relationship field), return a
    projector which returns a list of the PK of each item in the relationship (or
    just a single PK if this is a to-one field, but this is an inefficient way of
    doing it).
    """
    return relationship(name, attrgetter("pk"))


def combine(*projectors):
    """
    Given a list of projectors as *args, return another projector which calls each
    projector in turn and merges the resulting dictionaries.
    """

    def combined(instance):
        result = {}
        for projector in projectors:
            result.update(projector(instance))
        return result

    return combined


def alias(alias_or_aliases, projector):
    """
    Given a projector and a dictionary of aliases {"old_key_name": "new_key_name"},
    return a projector which replaces the keys in the output of the original projector
    with those provided in the alias map. As a shortcut, the argument can be a single
    string, in which case this will automatically alias a single-key projector without
    needing to know the key name of key in the dictionary returned from the
    inner projector.
    """

    def aliaser(instance):
        projected = projector(instance)
        if isinstance(alias_or_aliases, str) and len(projected) != 1:
            raise TypeError(
                "A single string can only be used as an alias for projectors "
                "that return a dictionary with a single key. Please use a mapping "
                "to define aliases instead."
            )

        alias_map = (
            {next(iter(projected)): alias_or_aliases}
            if isinstance(alias_or_aliases, str)
            else alias_or_aliases
        )
        for old, new in alias_map.items():
            projected[new] = projected.pop(old)
        return projected

    return aliaser


def noop(instance):
    return {}



from django_readers import projectors, qs


def field(name, *, transform_value=None, transform_value_if_none=False):
    return qs.include_fields(name), projectors.attr(
        name,
        transform_value=transform_value,
        transform_value_if_none=transform_value_if_none,
    )


def combine(*pairs):
    """
    Given a list of pairs as *args, return a pair which pipes together the prepare
    function from each pair, and combines the project function from each pair.
    """
    prepare_fns, project_fns = zip(*pairs)
    return qs.pipe(*prepare_fns), projectors.combine(*project_fns)


def alias(alias_or_aliases, pair):
    prepare, project = pair
    return prepare, projectors.alias(alias_or_aliases, project)


def prepare_only(prepare):
    return prepare, projectors.noop


def project_only(project):
    return qs.noop, project


def field_display(name):
    """
    Works with Django's get_FOO_display mechanism for fields with choices set. Given
    the name of a field, calls get_<name>_display, and returns a projector that puts
    the returned value under the key <name>_display.
    """
    return qs.include_fields(name), projectors.alias(
        f"{name}_display", projectors.method(f"get_{name}_display")
    )


def filter(*args, **kwargs):
    return prepare_only(qs.filter(*args, **kwargs))


def exclude(*args, **kwargs):
    return prepare_only(qs.exclude(*args, **kwargs))


def order_by(*args, **kwargs):
    return prepare_only(qs.order_by(*args, **kwargs))


"""
Below are pair functions which return the various queryset functions that prefetch
relationships of various types, and then project those related objects.

There are functions for forward, reverse or many-to-many relationships, and then
a `relationship` function which selects the correct one by introspecting the
model. It shouldn't usually be necessary to use the manual functions unless you're
doing something weird, like providing a custom queryset.
"""


def forward_relationship(name, related_queryset, relationship_pair, to_attr=None):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.prefetch_forward_relationship(
        name, related_queryset, prepare_related_queryset, to_attr
    )
    return prepare, projectors.relationship(to_attr or name, project_relationship)


def reverse_relationship(
    name, related_name, related_queryset, relationship_pair, to_attr=None
):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.prefetch_reverse_relationship(
        name, related_name, related_queryset, prepare_related_queryset, to_attr
    )
    return prepare, projectors.relationship(to_attr or name, project_relationship)


def many_to_many_relationship(name, related_queryset, relationship_pair, to_attr=None):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.prefetch_many_to_many_relationship(
        name, related_queryset, prepare_related_queryset, to_attr
    )
    return prepare, projectors.relationship(to_attr or name, project_relationship)


def relationship(name, relationship_pair, to_attr=None):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.auto_prefetch_relationship(name, prepare_related_queryset, to_attr)
    return prepare, projectors.relationship(to_attr or name, project_relationship)


def pk_list(name, to_attr=None):
    return (
        qs.auto_prefetch_relationship(name, qs.include_fields("pk"), to_attr=to_attr),
        projectors.pk_list(to_attr or name),
    )



try:
    import zen_queries
except ImportError:
    zen_queries = None


def map_or_apply(obj, fn):
    """
    If the first argument is iterable, map the function across each item in it and
    return the result. If it looks like a queryset or manager, call `.all()` and
    map the function across the result of that. If it's is a single item, just call
    the function on that item and return the result.
    """
    if obj is None:
        return None

    try:
        # Is the object itself iterable?
        return [fn(item) for item in iter(obj)]
    except TypeError:
        try:
            # Does the object have a `.all()` method (is it a manager?)
            return [fn(item) for item in obj.all()]
        except AttributeError:
            # It must be a single object
            return fn(obj)


def queries_disabled(pair):
    prepare, project = pair
    decorator = zen_queries.queries_disabled() if zen_queries else lambda fn: fn
    return decorator(prepare), decorator(project)



from django_readers import pairs
from django_readers.utils import queries_disabled


def process_item(item):
    if isinstance(item, str):
        return pairs.field(item)
    if isinstance(item, dict):
        return pairs.combine(
            *[
                relationship(name, relationship_spec)
                for name, relationship_spec in item.items()
            ]
        )
    return item


def process(spec):
    return queries_disabled(pairs.combine(*(process_item(item) for item in spec)))


def alias(alias_or_aliases, item):
    return pairs.alias(alias_or_aliases, process_item(item))


def relationship(name, relationship_spec, to_attr=None):
    return pairs.relationship(name, process(relationship_spec), to_attr)
